
#include "PathPlanningWindow.h"
#include "Planner/PFPlanner.h"
#include "Planner/RRTPlanner.h"
#include "Planner/AStarPlanner.h"
#include "VirtualRobot/EndEffector/EndEffector.h"
#include "VirtualRobot/Workspace/Reachability.h"
#include <VirtualRobot/RuntimeEnvironment.h>
#include <VirtualRobot/XML/SceneIO.h>

#include <QFileDialog>
#include <Eigen/Geometry>

#include <time.h>
#include <vector>
#include <iostream>
#include <cmath>

#include "Inventor/actions/SoLineHighlightRenderAction.h"
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoUnits.h>

#include <sstream>
using namespace std;
using namespace VirtualRobot;

float TIMER_MS = 30.0f;

PathPlanningWindow::PathPlanningWindow(Qt::WFlags flags)
:QMainWindow(NULL)
{
	VR_INFO << " start " << endl;
	//resize(1100, 768);
    robotFilename = "robots/ArmarIII/ArmarIII.xml";
    VirtualRobot::RuntimeEnvironment::getDataFileAbsolute(robotFilename);
   
	sceneSep = new SoSeparator;
	sceneSep->ref();
    robotSep = new SoSeparator;
    sceneSep->addChild(robotSep);
    obstacleSep = new SoSeparator;
    sceneSep->addChild(obstacleSep);
    floorSep = new SoSeparator;
    sceneSep->addChild(floorSep);
    trajectorySep = new SoSeparator;
    sceneSep->addChild(trajectorySep);

    //Eigen::Vector3f floorPos = Eigen::Vector3f::Zero();
    //Eigen::Vector3f floorUp = Eigen::Vector3f::UnitZ();
    //float floorExtendMM = 20000.0f;
    //SoNode * n = (SoNode*)CoinVisualizationFactory::CreatePlaneVisualization(floorPos,floorUp,floorExtendMM,0.0f);


	setupUI();

    std::string sceneFile1(DEMO_BASE_DIR "/data/PathPlanningScene1.xml");
    std::string sceneFile2(DEMO_BASE_DIR "/data/PathPlanningScene2.xml");
    std::string sceneFile3(DEMO_BASE_DIR "/data/PathPlanningScene3.xml");
    sceneFiles.push_back(sceneFile1);
    sceneFiles.push_back(sceneFile2);
    sceneFiles.push_back(sceneFile3);

    //setup box scene
    UI.comboBoxScene->setCurrentIndex(1);
    comboBoxSelectScene(1);

    //setup pf planner
    UI.comboBoxAlgo->setCurrentIndex(2);
    comboBoxSelectAlgorithm(2);

	viewer->viewAll();
}

bool PathPlanningWindow::loadRobot(std::string robotFilename)
{
    cout << "Loading Robot from " << robotFilename << endl;

    try
    {
        robot = RobotIO::loadRobot(robotFilename,RobotIO::eFull);
    }
    catch (VirtualRobotException &e)
    {
        cout << " ERROR while creating robot" << endl;
        cout << e.what();
        return false;
    }

    if (!robot)
    {
        cout << " ERROR while creating robot" << endl;
        return false;
    }
    try
    {
        VirtualRobot::BoundingBox bbox = robot->getBoundingBox();

        Eigen::Matrix4f gp = Eigen::Matrix4f::Identity();
        //gp(2,3) = 5.0f;
        gp(2,3) = -bbox.getMin()(2) + 4.0f;
        gp(0,3) = 2500;
        gp(1,3) = 1000;
        robot->setGlobalPose(gp);
        
    }catch (VirtualRobotException &e)
    {
        cout << " ERROR while building dynamic robot" << endl;
        cout << e.what();
        return false;
    }
//    updateJoints();
    return true;
}


PathPlanningWindow::~PathPlanningWindow()
{
	robot.reset();
	sceneSep->unref();
}

/*
void CPathPlanningWindow::saveScreenshot()
{
	static int counter = 0;
	SbString framefile;

	framefile.sprintf("MPL_Render_Frame%06d.png", counter);
	counter++;

	viewer->getSceneManager()->render();
	viewer->getSceneManager()->scheduleRedraw();
	QGLWidget* w = (QGLWidget*)viewer->getGLWidget();

	QImage i = w->grabFrameBuffer();
	bool bRes = i.save(framefile.getString(), "PNG");
	if (bRes)
		cout << "wrote image " << counter << endl;
	else
		cout << "failed writing image " << counter << endl;

}*/

void PathPlanningWindow::setupUI()
{
	 UI.setupUi(this);
	 //centralWidget()->setLayout(UI.gridLayoutViewer);
	viewer = new SoQtExaminerViewer(UI.frameViewer,"",TRUE,SoQtExaminerViewer::BUILD_POPUP);

	// setup
	viewer->setBackgroundColor(SbColor(1.0f, 1.0f, 1.0f));
	viewer->setAccumulationBuffer(true);
	viewer->setAntialiasing(true, 4);

	viewer->setGLRenderAction(new SoLineHighlightRenderAction);
	viewer->setTransparencyType(SoGLRenderAction::BLEND);
	viewer->setFeedbackVisibility(true);
	viewer->setSceneGraph(sceneSep);

    viewer->setAccumulationBuffer(true);
    viewer->setAntialiasing(true, 4);

	viewer->viewAll();

    connect(UI.pushButtonReset, SIGNAL(clicked()), this, SLOT(resetSceneryAll()));
    connect(UI.pushButtonRunPlanner, SIGNAL(clicked()), this, SLOT(buttonRunPlanner()));
    connect(UI.pushButtonStep1, SIGNAL(clicked()), this, SLOT(buttonStep1()));
    connect(UI.pushButtonStep2, SIGNAL(clicked()), this, SLOT(buttonStep2()));
    connect(UI.pushButtonStep3, SIGNAL(clicked()), this, SLOT(buttonStep3()));

    connect(UI.checkBoxColModel, SIGNAL(clicked()), this, SLOT(rebuildVisualization()));
    connect(UI.checkBoxRobot, SIGNAL(clicked()), this, SLOT(rebuildVisualization()));
    connect(UI.checkBoxFloor, SIGNAL(clicked()), this, SLOT(rebuildVisualization()));
    connect(UI.checkBoxObstacles, SIGNAL(clicked()), this, SLOT(rebuildVisualization()));
    connect(UI.checkBoxOption1, SIGNAL(clicked()), this, SLOT(checkBoxOption1()));

    QString scene1("Empty");
    QString scene2("Box");
    QString scene3("Trap");
    UI.comboBoxScene->addItem(scene1);
    UI.comboBoxScene->addItem(scene2);
    UI.comboBoxScene->addItem(scene3);

    QString algo1("PF Planner");
    QString algo2("RRT Planner");
    QString algo3("A* Planner");
    UI.comboBoxAlgo->addItem(algo1);
    UI.comboBoxAlgo->addItem(algo2);
    UI.comboBoxAlgo->addItem(algo3);

    connect(UI.comboBoxScene, SIGNAL(activated(int)), this, SLOT(comboBoxSelectScene(int)));
    connect(UI.comboBoxAlgo, SIGNAL(activated(int)), this, SLOT(comboBoxSelectAlgorithm(int)));


    connect(UI.horizontalSliderTrajectory, SIGNAL(valueChanged(int)), this, SLOT(sliderTrajectory(int)));

}


void PathPlanningWindow::resetSceneryAll()
{
	if (!robot)
		return;
    /*RobotConfigPtr c = robot->getConfig();

    c->setJointValues()
	std::vector<float> jv(allRobotNodes.size(),0.0f);
    robot->setJointValues(allRobotNodes,jv);*/

    comboBoxSelectScene(UI.comboBoxScene->currentIndex());
}



void PathPlanningWindow::checkBoxOption1()
{
    cout << "option 1 clicked" << endl;
}

void PathPlanningWindow::rebuildVisualization()
{
    SceneObject::VisualizationType colModel = (UI.checkBoxColModel->isChecked())?SceneObject::Collision:SceneObject::Full;

    robotSep->removeAllChildren();

    if (robot && UI.checkBoxRobot->isChecked())
    {
        boost::shared_ptr<CoinVisualization> visualization = robot->getVisualization<CoinVisualization>(colModel);
        SoNode* visualisationNode = NULL;
        if (visualization)
            visualisationNode = visualization->getCoinVisualization();

        if (visualisationNode)
            robotSep->addChild(visualisationNode);
    }

    obstacleSep->removeAllChildren();
    if (obstacles && UI.checkBoxObstacles->isChecked())
    {

        for (size_t i=0;i<obstacles->getSceneObjects().size();i++)
        {
            boost::shared_ptr<CoinVisualization> visualization = obstacles->getSceneObjects().at(i)->getVisualization<CoinVisualization>(colModel);
            SoNode* visualisationNode = NULL;
            if (visualization)
                visualisationNode = visualization->getCoinVisualization();

            if (visualisationNode)
                obstacleSep->addChild(visualisationNode);
        }
    }

    floorSep->removeAllChildren();
    if (floor && UI.checkBoxFloor->isChecked())
    {

        boost::shared_ptr<CoinVisualization> visualization = floor->getVisualization<CoinVisualization>(colModel);
        SoNode* visualisationNode = NULL;
        if (visualization)
            visualisationNode = visualization->getCoinVisualization();

        if (visualisationNode)
            floorSep->addChild(visualisationNode);
    }

}


void PathPlanningWindow::closeEvent(QCloseEvent *event)
{
	quit();
	QMainWindow::closeEvent(event);
}



int PathPlanningWindow::main()
{
	SoQt::show(this);
	SoQt::mainLoop();
	return 0;
}


void PathPlanningWindow::quit()
{
    std::cout << "PathPlanningWindow: Closing" << std::endl;
	this->close();
	SoQt::exitMainLoop();
}

void PathPlanningWindow::buttonRunPlanner()
{
    if (!robot) 
        return;
    Eigen::Matrix4f gp = robot->getGlobalPose();
    Eigen::Vector2f pos(gp(0,3), gp(1,3));
    Eigen::Vector2f goal(2500, 9000);
    std::vector<Eigen::Vector2f> result = p->plan(pos, goal);
    trajectory = result;
    showTrajectory(trajectory);

}

void PathPlanningWindow::buttonStep1()
{
    cout << "Step 1..." << endl;
}

void PathPlanningWindow::buttonStep2()
{
    cout << "Step 2..." << endl;
}

void PathPlanningWindow::buttonStep3()
{
    cout << "Step 3..." << endl;
}

void PathPlanningWindow::showTrajectory(std::vector< Eigen::Vector2f > &traj)
{
    trajectorySep->removeAllChildren();
    SoSeparator *s = new SoSeparator;
    trajectorySep->addChild(s);
    SoUnits *u = new SoUnits;
    u->units = SoUnits::MILLIMETERS;
    s->addChild(u);
    SoSphere *sphere = new SoSphere;
    sphere->radius = 10.0f;
    Eigen::Matrix4f m = Eigen::Matrix4f::Identity();
    Eigen::Matrix4f mold = Eigen::Matrix4f::Identity();
    
    for (size_t i=0;i<traj.size();i++)
    {
        SoSeparator *s2 = new SoSeparator;
        m.block(0,3,2,1) = traj[i];
        s2->addChild(CoinVisualizationFactory::getMatrixTransform(m));
        s2->addChild(sphere);
        s->addChild(s2);
        if (i>0)
        {
            s->addChild(CoinVisualizationFactory::createCoinLine(mold,m,1.5f,0,0,1.f));
        }
        mold = m;
    }
}

void PathPlanningWindow::sliderTrajectory(int pos)
{	

    float fPos = (float)pos / 10.0f;

    // update trajectory
    if (robot && trajectory.size()>0)
    {
        int trajNr = int(trajectory.size() * fPos/100.0f);
        Eigen::Matrix4f m = Eigen::Matrix4f::Identity();
        if (trajNr>=int(trajectory.size()))
            trajNr = trajectory.size()-1;
        m.block(0,3,2,1) = trajectory[trajNr];
        robot->setGlobalPose(m);
    }
    //


    UI.lcdNumberTrajectory->display((double)fPos);

}


void PathPlanningWindow::comboBoxSelectScene(int scene)
{
    cout << "Loading Scene nr " << (scene+1) << ". File:" << sceneFiles[scene] << endl;
    loadScene(sceneFiles[scene]);
    // (re-) setup planner
    comboBoxSelectAlgorithm(UI.comboBoxAlgo->currentIndex());
}

void PathPlanningWindow::comboBoxSelectAlgorithm(int algo)
{
    if (!robot)
        return;

    cout << "Selecting algorithm: " << algo << endl;
    if (algo == 0) {
        p.reset(new PFPlanner(robot,obstacles,100.f, 1.f, 10000.f, 500.0f));
    } else if (algo == 1) {
        p.reset(new RRTPlanner(robot,obstacles, 100.f, 50.f, 0.01f));
    } else if (algo == 2) {
        p.reset(new AStarPlanner(robot, obstacles, 100.f));
    }
    p->setRobotColModel("Platform");
}

void PathPlanningWindow::loadScene(std::string sceneFile)
{
	robotSep->removeAllChildren();
    cout << "Loading Scene from " << sceneFile << endl;
    scene.reset();
    robot.reset();

	try
	{
        scene = SceneIO::loadScene(sceneFile);
	}
	catch (VirtualRobotException &e)
	{
        cout << " ERROR while creating scene" << endl;
		cout << e.what();
		return;
	}
	
    if (!scene)
	{
        cout << " ERROR while creating scene" << endl;
		return;
	}
    loadRobot(robotFilename);
    scene->registerRobot(robot);

    obstacles = scene->getSceneObjectSet("Obstacles");

    if (!obstacles)
        cout << "No SceneObjectSet with name 'Obstacles' found in scene definition!!!" << endl;
    floor = scene->getObstacle("Floor");

	// build visualization
	rebuildVisualization();
	viewer->viewAll();

    // reset trajectory
    trajectory.clear();
    showTrajectory(trajectory);
}


